# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import unittest

from sudachipy.dictionarylib import charactercategory
from sudachipy.dictionarylib.categorytype import CategoryType


class TestCharacterCategory(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.test_resources_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.pardir,
            'resources')
        pass

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_range_containing_length(self):
        range_ = charactercategory.CharacterCategory.Range()
        range_.low = 0x41
        range_.high = 0x54
        self.assertEqual(3, range_.containing_length('ABC12'))
        self.assertEqual(0, range_.containing_length('熙'))

    def test_get_category_types(self):
        cat = charactercategory.CharacterCategory()
        cat.read_character_definition(os.path.join(self.test_resources_dir, 'char.def'))
        self.assertEqual({CategoryType.KANJI}, cat.get_category_types(ord('熙')))
        self.assertNotEqual({CategoryType.DEFAULT}, cat.get_category_types(ord('熙')))

    def test_read_character_definition(self):
        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("#\n \n")
            wf.write("0x0030..0x0039 NUMERIC\n")
            wf.write("0x0032         KANJI\n")
        cat = charactercategory.CharacterCategory()
        cat.read_character_definition(f)
        self.assertEqual({CategoryType.NUMERIC}, cat.get_category_types(0x0030))
        self.assertEqual({CategoryType.NUMERIC}, cat.get_category_types(0x0031))
        self.assertEqual({CategoryType.NUMERIC, CategoryType.KANJI}, cat.get_category_types(0x0032))
        self.assertEqual({CategoryType.NUMERIC}, cat.get_category_types(0x0033))
        self.assertEqual({CategoryType.NUMERIC}, cat.get_category_types(0x0039))

        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("#\n \n")
            wf.write("0x0030..0x0039 NUMERIC\n")
            wf.write("0x0070..0x0079 ALPHA\n")
            wf.write("0x3007         KANJI\n")
            wf.write("0x0030         KANJI\n")
        cat = charactercategory.CharacterCategory()
        cat.read_character_definition(f)
        self.assertEqual({CategoryType.NUMERIC, CategoryType.KANJI}, cat.get_category_types(0x0030))
        self.assertEqual({CategoryType.NUMERIC}, cat.get_category_types(0x0039))
        self.assertEqual({CategoryType.KANJI}, cat.get_category_types(0x3007))
        self.assertEqual({CategoryType.DEFAULT}, cat.get_category_types(0x0069))
        self.assertEqual({CategoryType.ALPHA}, cat.get_category_types(0x0070))
        self.assertEqual({CategoryType.DEFAULT}, cat.get_category_types(0x0080))

        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("#\n \n")
            wf.write("0x0030..0x0039 KATAKANA\n")
            wf.write("0x3007         KANJI KANJINUMERIC\n")
            wf.write("0x3008         KANJI KANJINUMERIC\n")
            wf.write("0x3009         KANJI KANJINUMERIC\n")
            wf.write("0x0039..0x0040 ALPHA\n")
            wf.write("0x0030..0x0039 NUMERIC\n")
            wf.write("0x0030         KANJI\n")
        cat = charactercategory.CharacterCategory()
        cat.read_character_definition(f)
        self.assertEqual({CategoryType.DEFAULT}, cat.get_category_types(0x0029))
        self.assertEqual({CategoryType.NUMERIC, CategoryType.KANJI, CategoryType.KATAKANA}, cat.get_category_types(0x0030))
        self.assertEqual({CategoryType.NUMERIC, CategoryType.KATAKANA, CategoryType.ALPHA}, cat.get_category_types(0x0039))
        self.assertEqual({CategoryType.ALPHA}, cat.get_category_types(0x0040))
        self.assertEqual({CategoryType.DEFAULT}, cat.get_category_types(0x0041))
        self.assertEqual({CategoryType.KANJI, CategoryType.KANJINUMERIC}, cat.get_category_types(0x3007))
        self.assertEqual({CategoryType.DEFAULT}, cat.get_category_types(0x4007))

        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("0x4E00..0x9FFF  KANJI\n")
            wf.write("0x4E8C KANJINUMERIC KANJI\n")
        cat = charactercategory.CharacterCategory()
        cat.read_character_definition(f)
        self.assertEqual({CategoryType.KANJI}, cat.get_category_types(ord('男')))
        self.assertEqual({CategoryType.KANJI}, cat.get_category_types(0x4E8B))
        self.assertEqual({CategoryType.KANJI, CategoryType.KANJINUMERIC}, cat.get_category_types(0x4E8C))
        self.assertEqual({CategoryType.KANJI}, cat.get_category_types(0x4E8D))

    def test_read_character_definition_with_invalid_format(self):
        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("0x0030..0x0039\n")
        cat = charactercategory.CharacterCategory()
        with self.assertRaises(AttributeError) as cm:
            cat.read_character_definition(f)
        self.assertEqual('invalid format at line 0', cm.exception.args[0])

    def test_read_character_definition_with_invalid_range(self):
        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("0x0030..0x0029 NUMERIC\n")
        cat = charactercategory.CharacterCategory()
        with self.assertRaises(AttributeError) as cm:
            cat.read_character_definition(f)
        self.assertEqual('invalid range at line 0', cm.exception.args[0])

    def test_read_character_definition_with_invalid_type(self):
        f = os.path.join(self.test_dir, 'test_file.txt')
        with open(f, 'w') as wf:
            wf.write("0x0030..0x0039 FOO\n")
        cat = charactercategory.CharacterCategory()
        with self.assertRaises(AttributeError) as cm:
            cat.read_character_definition(f)
        self.assertEqual('FOO is invalid type at line 0', cm.exception.args[0])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from sudachipy.dictionarylib.charactercategory import CharacterCategory
from sudachipy.dictionarylib.grammar import Grammar

mocked_grammar = mock.Mock(spec=Grammar)
mocked_grammar.get_part_of_speech_size.return_value = 0
mocked_grammar.get_part_of_speech_string.return_value = None
mocked_grammar.get_part_of_speech_id.return_value = 0
mocked_grammar.get_connect_cost.return_value = 0
# mocked_grammar.set_connect_cost.return_value = None
mocked_grammar.get_bos_parameter.return_value = None
mocked_grammar.get_eos_parameter.return_value = None


def mocked_get_character_category():
    cat = CharacterCategory()
    test_resources_dir = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        os.pardir,
        'sudachipy',
        'resources')
    try:
        cat.read_character_definition(os.path.join(test_resources_dir, 'char.def'))
    except IOError as e:
        print(e)
    return cat


mocked_grammar.get_character_category.side_effect = mocked_get_character_category


mocked_grammar.set_character_category.return_value = None



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from sudachipy.dictionarylib.wordinfo import WordInfo


class PathRewritePlugin(ABC):

    @abstractmethod
    def set_up(self, grammar):
        raise NotImplementedError

    @abstractmethod
    def rewrite(self, text, path, lattice):
        raise NotImplementedError

    def concatenate(self, path, begin, end, lattice, normalized_form):
        if begin >= end:
            raise IndexError("begin >= end")
        b = path[begin].get_begin()
        e = path[end - 1].get_end()
        pos_id = path[begin].get_word_info().pos_id
        surface = ""
        length = 0
        normalized_builder, dictionary_builder, reading_builder = "", "", ""
        for i in range(begin, end):
            info = path[i].get_word_info()
            surface += info.surface
            length += info.head_word_length
            if not normalized_form:
                normalized_builder += info.normalized_form
            dictionary_builder += info.dictionary_form
            reading_builder += info.reading_form

        normalized_form = normalized_form if normalized_form else normalized_builder
        wi = WordInfo(surface=surface, head_word_length=length, pos_id=pos_id,
                      normalized_form=normalized_form, dictionary_form=dictionary_builder, dictionary_form_word_id=-1,
                      reading_form=reading_builder, a_unit_split=[], b_unit_split=[], word_structure=[], synonym_group_ids=[])

        node = lattice.create_node()
        node.set_range(b, e)
        node.set_word_info(wi)

        path[begin:end] = [node]
        return node

    def concatenate_oov(self, path, begin, end, pos_id, lattice):
        if begin >= end:
            raise IndexError("begin >= end")
        b = path[begin].get_begin()
        e = path[end - 1].get_end()
        surface = ""
        length = 0
        for i in range(begin, end):
            info = path[i].get_word_info()
            surface += info.surface
            length += info.head_word_length

        wi = WordInfo(surface=surface, head_word_length=length, pos_id=pos_id,
                      normalized_form=surface, dictionary_form=surface, dictionary_form_word_id=-1,
                      reading_form="", a_unit_split=[], b_unit_split=[], word_structure=[], synonym_group_ids=[])

        node = lattice.create_node()
        node.set_range(b, e)
        node.set_word_info(wi)

        path[begin:end] = [node]
        return node

    def get_char_category_types(self, text, node):
        return text.get_char_category_types(node.get_begin(), node.get_end())



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import warnings
from importlib import import_module
from importlib.util import find_spec
from pathlib import Path
from typing import List

DEFAULT_RESOURCEDIR = Path(__file__).absolute().parent / 'resources'
DEFAULT_SETTINGFILE = DEFAULT_RESOURCEDIR / 'sudachi.json'
DEFAULT_RESOURCEDIR = DEFAULT_RESOURCEDIR.as_posix()
DEFAULT_SETTINGFILE = DEFAULT_SETTINGFILE.as_posix()


def get_absolute_dict_path(dict_type: str) -> str:
    pkg_path = Path(import_module('sudachidict_' + dict_type).__file__).parent
    dic_path = pkg_path / 'resources' / 'system.dic'
    return str(dic_path.absolute())


def to_absolute_resource_path(resource_dir: str, dict_path: str) -> str:
    if Path(dict_path).is_absolute():
        return dict_path
    else:
        return os.path.join(resource_dir, dict_path)


def find_dict_path(dict_type='core'):
    is_installed = find_spec('sudachidict_{}'.format(dict_type))
    if is_installed:
        return get_absolute_dict_path(dict_type)
    else:
        raise ModuleNotFoundError(
            'Package `sudachidict_{}` dose not exist. '
            'You may install it with a command `$ pip install sudachidict_{}`'.format(dict_type, dict_type)
        )


class _Settings(object):

    DICT_PATH_KEY = 'systemDict'
    CHAR_DEF_KEY = 'characterDefinitionFile'
    USER_DICT_PATH_KEY = 'userDict'

    def __init__(self):
        self.__is_active = False
        self.__dict_ = None
        self.__config_path = None
        self.resource_dir = None

    def set_up(self, config_path=None, resource_dir=None, dict_type=None) -> None:
        config_path = config_path or DEFAULT_SETTINGFILE
        self.__config_path = config_path
        resource_dir = resource_dir or os.path.dirname(config_path)
        with open(config_path, 'r', encoding='utf-8') as f:
            self.__dict_ = json.load(f)
        self.__is_active = True
        self.resource_dir = resource_dir
        if dict_type is not None:
            if dict_type in ['small', 'core', 'full']:
                if self.DICT_PATH_KEY in self.__dict_ and self.__dict_[self.DICT_PATH_KEY] and \
                        'sudachidict_{}'.format(dict_type) not in self.__dict_[self.DICT_PATH_KEY]:
                    warnings.warn(
                        'Two system dictionaries may be specified. '
                        'The `sudachidict_{}` defined "dict_type" overrides those defined in the config file.'.format(dict_type)
                    )
                self.__dict_[self.DICT_PATH_KEY] = find_dict_path(dict_type=dict_type)
            else:
                raise ValueError('"dict_type" must be "small", "core", or "full".')
        else:
            if self.DICT_PATH_KEY not in self.__dict_ or not self.__dict_[self.DICT_PATH_KEY]:
                self.__dict_[self.DICT_PATH_KEY] = find_dict_path()

    def __setitem__(self, key, value):
        if not self.__is_active:
            self.set_up()
        self.__dict_[key] = value

    def __getitem__(self, key):
        if not self.__is_active:
            self.set_up()
        return self.__dict_[key]

    def keys(self):
        return self.__dict_.keys()

    def __contains__(self, item):
        return item in self.__dict_.keys()

    def system_dict_path(self) -> str:
        dict_path = self.__dict_[self.DICT_PATH_KEY]
        return to_absolute_resource_path(self.resource_dir, dict_path)

    def char_def_path(self) -> str:
        key = self.CHAR_DEF_KEY
        if key in self.__dict_:
            return to_absolute_resource_path(self.resource_dir, self.__dict_[key])
        raise KeyError('`{}` not defined in setting file'.format(key))

    def user_dict_paths(self) -> List[str]:
        key = self.USER_DICT_PATH_KEY
        if key in self.__dict_:
            return [to_absolute_resource_path(self.resource_dir, path) for path in self.__dict_[key]]
        return []


settings = _Settings()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import time
from logging import getLogger
from unittest import TestCase

from sudachipy.dictionarylib import SYSTEM_DICT_VERSION_2
from sudachipy.dictionarylib.dictionaryheader import DictionaryHeader
from sudachipy.dictionarylib.userdictionarybuilder import UserDictionaryBuilder

from .test_dictionarybuilder import TestDictionaryBuilder


class TestUserDictionaryBuilder(TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        test_resources_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.pardir,
            'resources')
        self.dict_filename = os.path.join(test_resources_dir, 'system.dic')
        _, _, self.grammar, self.lexicon_set = \
            TestDictionaryBuilder.read_system_dictionary(self.dict_filename)
        self.logger = getLogger()
        self.logger.disabled = True

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_parseline_with_userdefined_POS(self):
        builder = UserDictionaryBuilder(self.grammar, self.lexicon_set, logger=self.logger)
        builder.parse_line('田中,0,0,0,田中,存在,しない,品詞,*,*,*,タナカ,田中,*,A,*,*,*,*'.split(','))
        self.assertEqual(1, len(builder.pos_table.get_list()))

    def test_build(self):
        out_path = os.path.join(self.test_dir, 'output.txt')
        in_path = os.path.join(self.test_dir, 'input.txt')

        out_stream = open(out_path, 'wb')
        # lexicon_paths = [self.input_path]
        # matrix_input_stream = open(self.matrix_path, 'r')
        with open(in_path, 'w', encoding='utf-8') as wf:
            wf.write("東京都市,0,0,0,東京都市,名詞,固有名詞,地名,一般,*,*,ヒガシキョウトシ,東京都市,*,B,\"東,名詞,普通名詞,一般,*,*,*,ヒガシ/3/U1\",*,\"4/3/市,名詞,普通名詞,一般,*,*,*,シ\",*\n")
            wf.write('市,-1,-1,0,市,名詞,普通名詞,一般,*,*,*,シ,市,*,A,*,*,*,*\n')

        _, _, grammar, lexicon_set = TestDictionaryBuilder.read_system_dictionary(self.dict_filename)
        header = DictionaryHeader(SYSTEM_DICT_VERSION_2, int(time.time()), 'test')
        out_stream.write(header.to_bytes())
        builder = UserDictionaryBuilder(grammar, lexicon_set, logger=self.logger)
        lexicon_paths = [in_path]
        builder.build(lexicon_paths, None, out_stream)
        out_stream.close()

        buffers, header, grammar, lexicon_set = TestDictionaryBuilder.read_system_dictionary(out_path)
        lexicon = lexicon_set.lexicons[0]

        # header
        self.assertEqual(SYSTEM_DICT_VERSION_2, header.version)
        self.assertEqual('test', header.description)

        # lexicon
        self.assertEqual(0, lexicon.get_left_id(0))
        self.assertEqual(0, lexicon.get_cost(0))
        wi = lexicon.get_word_info(0)
        self.assertEqual('東京都市', wi.surface)
        self.assertEqual('東京都市', wi.normalized_form)
        self.assertEqual(-1, wi.dictionary_form_word_id)
        self.assertEqual('ヒガシキョウトシ', wi.reading_form)
        self.assertEqual(3, wi.pos_id)
        self.assertEqual([4, 3, 1 | (1 << 28)], wi.a_unit_split)
        self.assertEqual([], wi.b_unit_split)
        self.assertEqual([4, 3, 1 | (1 << 28)], wi.word_structure)
        lst = lexicon.lookup('東京都市'.encode('utf-8'), 0)
        self.assertEqual((0, len('東京都市'.encode('utf-8'))), lst.__next__())
        with self.assertRaises(StopIteration):
            lst.__next__()

        self.assertEqual(-1, lexicon.get_left_id(1))
        self.assertEqual(0, lexicon.get_cost(1))
        wi = lexicon.get_word_info(1)
        self.assertEqual('市', wi.surface)
        self.assertEqual('市', wi.normalized_form)
        self.assertEqual(-1, wi.dictionary_form_word_id)
        self.assertEqual('シ', wi.reading_form)
        self.assertEqual(4, wi.pos_id)
        self.assertEqual([], wi.a_unit_split)
        self.assertEqual([], wi.b_unit_split)
        lst = lexicon.lookup('東'.encode('utf-8'), 0)
        with self.assertRaises(StopIteration):
            lst.__next__()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

from distutils.extension import Extension

extensions = [
        Extension('sudachipy.latticenode', ['sudachipy/latticenode.pyx']),
        Extension('sudachipy.lattice', ['sudachipy/lattice.pyx']),
        Extension('sudachipy.tokenizer', ['sudachipy/tokenizer.pyx']),
        ]

setup(name="SudachiPy",
      use_scm_version=True,
      setup_requires=['setuptools_scm', 'cython'],
      description="Python version of Sudachi, the Japanese Morphological Analyzer",
      long_description=open('README.md', encoding='utf-8').read(),
      long_description_content_type="text/markdown",
      url="https://github.com/WorksApplications/SudachiPy",
      license="Apache-2.0",
      author="Works Applications",
      author_email="sudachi@worksap.co.jp",
      packages=find_packages(include=["sudachipy", "sudachipy.*"]),
      package_data={"": ["resources/*.json", "resources/*.dic", "resources/*.def"]},
      entry_points={
          "console_scripts": ["sudachipy=sudachipy.command_line:main"],
      },
      install_requires=[
            "sortedcontainers~=2.1.0",
            'dartsclone~=0.9.0',
      ],
      ext_modules=extensions,
      )



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mmap
import os
import unittest

from sudachipy.dictionarylib import SYSTEM_DICT_VERSION_2
from sudachipy.dictionarylib.dictionaryheader import DictionaryHeader


class TestDictionaryHeader(unittest.TestCase):

    def setUp(self):
        # Copied from sudachipy.dictionay.Dictionary.read_system_dictionary
        test_resources_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.pardir,
            'resources')
        filename = os.path.join(test_resources_dir, 'system.dic')
        with open(filename, 'rb') as system_dic:
            bytes_ = mmap.mmap(system_dic.fileno(), 0, access=mmap.ACCESS_READ)
        offset = 0
        self.header = DictionaryHeader.from_bytes(bytes_, offset)

    def test_version(self):
        self.assertEqual(SYSTEM_DICT_VERSION_2, self.header.version)

    def test_create_time(self):
        self.assertTrue(self.header.create_time > 0)

    def test_description(self):
        self.assertEqual("the system dictionary for the unit tests", self.header.description)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from sudachipy.plugin.path_rewrite.numericparser import NumericParser


class TestNumericParser(TestCase):

    def setUp(self) -> None:
        self.parser = NumericParser()

    def test_digits(self):
        self.assertTrue(self.parse('1000'))
        self.assertEqual('1000', self.parser.get_normalized())
        self.parser.clear()

    def test_starts_with_zero(self):
        self.assertTrue(self.parse('001000'))
        self.assertEqual('001000', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('〇一〇〇〇'))
        self.assertEqual('01000', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('00.1000'))
        self.assertEqual('00.1', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('000'))
        self.assertEqual('000', self.parser.get_normalized())
        self.parser.clear()

    def test_use_small_unit(self):
        self.assertTrue(self.parse('二十七'))
        self.assertEqual('27', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('千三百二十七'))
        self.assertEqual('1327', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('千十七'))
        self.assertEqual('1017', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('千三百二十七.〇五'))
        self.assertEqual('1327.05', self.parser.get_normalized())
        self.parser.clear()

        self.assertFalse(self.parse('三百二十百'))
        self.parser.clear()

    def test_use_large_unit(self):
        self.assertTrue(self.parse('1万'))
        self.assertEqual('10000', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('千三百二十七万'))
        self.assertEqual('13270000', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('千三百二十七万一四'))
        self.assertEqual('13270014', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('千三百二十七万一四.〇五'))
        self.assertEqual('13270014.05', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('三兆2千億千三百二十七万一四.〇五'))
        self.assertEqual('3200013270014.05', self.parser.get_normalized())
        self.parser.clear()

        self.assertFalse(self.parse('億万'))
        self.parser.clear()

    def test_float_with_unit(self):
        self.assertTrue(self.parse('1.5千'))
        self.assertEqual('1500', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('1.5百万'))
        self.assertEqual('1500000', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('1.5百万1.5千20'))
        self.assertEqual('1501520', self.parser.get_normalized())
        self.parser.clear()

        self.assertFalse(self.parse('1.5千5百'))
        self.parser.clear()

        self.assertFalse(self.parse('1.5千500'))
        self.parser.clear()

    def test_log_numeric(self):
        self.assertTrue(self.parse('200000000000000000000万'))
        self.assertEqual('2000000000000000000000000', self.parser.get_normalized())
        self.parser.clear()

    def test_with_comma(self):
        self.assertTrue(self.parse('2,000,000'))
        self.assertEqual('2000000', self.parser.get_normalized())
        self.parser.clear()

        self.assertTrue(self.parse('259万2,300'))
        self.assertEqual('2592300', self.parser.get_normalized())
        self.parser.clear()

        self.assertFalse(self.parse('200,00,000'))
        self.assertEqual(NumericParser.Error.COMMA, self.parser._error_state)
        self.parser.clear()

        self.assertFalse(self.parse('2,4'))
        self.assertEqual(NumericParser.Error.COMMA, self.parser._error_state)
        self.parser.clear()

        self.assertFalse(self.parse('000,000'))
        self.assertEqual(NumericParser.Error.COMMA, self.parser._error_state)
        self.parser.clear()

        self.assertFalse(self.parse(',000'))
        self.assertEqual(NumericParser.Error.COMMA, self.parser._error_state)
        self.parser.clear()

        self.assertFalse(self.parse('256,55.1'))
        self.assertEqual(NumericParser.Error.COMMA, self.parser._error_state)
        self.parser.clear()

    def test_not_digit(self):
        self.assertFalse(self.parse('@@@'))
        self.parser.clear()

    def test_float_point(self):
        self.assertTrue(self.parse('6.0'))
        self.assertEqual('6', self.parser.get_normalized())
        self.parser.clear()

        self.assertFalse(self.parse('6.'))
        self.assertEqual(NumericParser.Error.POINT, self.parser.error_state)
        self.parser.clear()

        self.assertFalse(self.parse('1.2.3'))
        self.assertEqual(NumericParser.Error.POINT, self.parser.error_state)
        self.parser.clear()

    def parse(self, text: str) -> bool:
        for c in text:
            if not self.parser.append(c):
                return False
        return self.parser.done()



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from sudachipy.dictionarylib.categorytype import CategoryType

from .numericparser import NumericParser
from .path_rewrite_plugin import PathRewritePlugin


class JoinNumericPlugin(PathRewritePlugin):

    _numeric_pos_id = None
    _enable_normalize = True

    def __init__(self, json_obj):
        self._NUMERIC_POS = ['名詞', '数詞', '*', '*', '*', '*']
        if not json_obj:
            return
        if 'joinKanjiNumeric' in json_obj:
            warnings.warn('joinKanjiNumeric is already nonsense key', SyntaxWarning)
        if 'enableNormalize' in json_obj:
            self._enable_normalize = json_obj['enableNormalize']

    def set_up(self, grammar):
        self._numeric_pos_id = grammar.get_part_of_speech_id(self._NUMERIC_POS)

    def rewrite(self, text, path, lattice):
        begin_index = -1
        comma_as_digit = True
        period_as_digit = True
        parser = NumericParser()
        i = -1

        while i < len(path) - 1:
            i += 1
            node = path[i]
            types = self.get_char_category_types(text, node)
            s = node.get_word_info().normalized_form
            if CategoryType.NUMERIC in types or CategoryType.KANJINUMERIC in types or \
               (period_as_digit and s == '.') or (comma_as_digit and s == ','):

                if begin_index < 0:
                    parser.clear()
                    begin_index = i

                for c in s:
                    if not parser.append(c):
                        if begin_index >= 0:
                            if parser.error_state == NumericParser.Error.COMMA:
                                comma_as_digit = False
                                i = begin_index - 1
                            elif parser.error_state == NumericParser.Error.POINT:
                                period_as_digit = False
                                i = begin_index - 1
                            begin_index = -1
                        break
                continue

            if begin_index >= 0:
                if parser.done():
                    self._concat(path, begin_index, i, lattice, parser)
                    i = begin_index + 1
                else:
                    ss = path[i - 1].get_word_info().normalized_form
                    if (parser.error_state == NumericParser.Error.COMMA and ss == ',') or \
                       (parser.error_state == NumericParser.Error.POINT and ss == '.'):
                        self._concat(path, begin_index, i - 1, lattice, parser)
                        i = begin_index + 2
            begin_index = -1
            if not comma_as_digit and s != ',':
                comma_as_digit = True
            if not period_as_digit and s != '.':
                period_as_digit = True

        if begin_index >= 0:
            if parser.done():
                self._concat(path, begin_index, len(path), lattice, parser)
            else:
                ss = path[-1].get_word_info().normalized_form
                if (parser.error_state == NumericParser.Error.COMMA and ss == ',') or \
                   (parser.error_state == NumericParser.Error.POINT and ss == '.'):
                    self._concat(path, begin_index, len(path) - 1, lattice, parser)

    def _concat(self, path, begin, end, lattice, parser) -> None:
        if path[begin].get_word_info().pos_id != self._numeric_pos_id:
            return
        if self._enable_normalize:
            normalized_form = parser.get_normalized()
            if end - begin > 1 or normalized_form != path[begin].get_word_info().normalized_form:
                self.concatenate(path, begin, end, lattice, normalized_form)
            return
        if end - begin > 1:
            self.concatenate(path, begin, end, lattice, '')



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Grammar:
    INHIBITED_CONNECTION = 0x7fff

    def __init__(self, bytes_, offset):
        self._POS_DEPTH = 6
        self._BOS_PARAMETER = [0, 0, 0]
        self._EOS_PARAMETER = [0, 0, 0]

        self.char_category = None
        self.is_copied_connect_table = False

        original_offset = bytes_.tell()
        bytes_.seek(offset)
        pos_size = self.bytes_get_short(bytes_)
        self.pos_list = []
        for i in range(pos_size):
            pos = []
            for j in range(self._POS_DEPTH):
                pos.append(self.bytes_get_string(bytes_))
            self.pos_list.append(pos)
        left_id_size = self.bytes_get_short(bytes_)
        right_id_size = self.bytes_get_short(bytes_)
        connect_table_offset = bytes_.tell()

        self.storage_size = (connect_table_offset - offset) + 2 * left_id_size * right_id_size

        self._matrix_view = \
            memoryview(bytes_)[connect_table_offset: connect_table_offset + 2 * left_id_size * right_id_size]
        if left_id_size * right_id_size != 0:
            self._matrix_view = self._matrix_view.cast('h', shape=[left_id_size, right_id_size])
        bytes_.seek(original_offset)

    def get_storage_size(self):
        return self.storage_size

    def get_part_of_speech_size(self):
        return len(self.pos_list)

    def get_part_of_speech_string(self, pos_id):
        return self.pos_list[pos_id]

    def get_part_of_speech_id(self, pos):
        return self.pos_list.index(pos) if pos in self.pos_list else -1

    def get_connect_cost(self, left: int, right: int) -> int:
        """ Returns connection cost of nodes

        Args:
            left: right-ID of left node
            right: left-ID of right node

        Returns:
            cost of connection

        """
        return self._matrix_view[right, left]

    def set_connect_cost(self, left: int, right: int, cost: int) -> None:
        """ Sets connection cost of nodes

        Note: bytes_ must be ACCESS_COPY mode

        Args:
            left: right-ID of left node
            right: left-ID of right node
            cost: cost of connection

        """
        self._matrix_view[right, left] = cost

    def get_bos_parameter(self):
        return self._BOS_PARAMETER

    def get_eos_parameter(self):
        return self._EOS_PARAMETER

    def get_character_category(self):
        return self.char_category

    def set_character_category(self, char_category):
        self.char_category = char_category

    @staticmethod
    def bytes_get_string(bytes_):
        length = bytes_.read_byte()
        string = bytes_.read(2 * length)
        return string.decode('utf-16')

    @staticmethod
    def bytes_get_short(bytes_):
        return int.from_bytes(bytes_.read(2), 'little', signed=True)

    def add_pos_list(self, grammar):
        self.pos_list.extend(grammar.pos_list)



# Copyright (c) 2019 Works Applications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mmap
import os
import unittest

from sudachipy.dictionarylib.dictionaryheader import DictionaryHeader
from sudachipy.dictionarylib.doublearraylexicon import DoubleArrayLexicon


class TestDoubleArrayLexicon(unittest.TestCase):

    __GRAMMAR_SIZE = 470

    def setUp(self):
        # Copied from sudachipy.dictionay.Dictionary.read_system_dictionary
        test_resources_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            os.pardir,
            'resources')
        filename = os.path.join(test_resources_dir, 'system.dic')
        with open(filename, 'rb') as system_dic:
            bytes_ = mmap.mmap(system_dic.fileno(), 0, access=mmap.ACCESS_READ)
        header = DictionaryHeader.from_bytes(bytes_, 0)
        if not header.is_system_dictionary():
            raise Exception('invalid system dictionary')
        self.lexicon = DoubleArrayLexicon(bytes_, header.storage_size() + 470, True)

    def test_lookup(self):
        res = self.lexicon.lookup('東京都'.encode('utf-8'), 0)
        self.assertEqual((4, 3), res.__next__())  # 東

        self.assertEqual((5, 6), res.__next__())  # 東京

        self.assertEqual((6, 9), res.__next__())  # 東京都

        with self.assertRaises(StopIteration):
            res.__next__()

        res = self.lexicon.lookup('東京都に'.encode('utf-8'), 9)
        self.assertEqual((1, 12), res.__next__())  # に(接続助詞)
        self.assertEqual((2, 12), res.__next__())  # に(格助詞)
        with self.assertRaises(StopIteration):
            res.__next__()

        res = self.lexicon.lookup('あれ'.encode('utf-8'), 0)
        with self.assertRaises(StopIteration):
            res.__next__()

    def test_parameters(self):
        # た
        self.assertEqual(1, self.lexicon.get_left_id(0))
        self.assertEqual(1, self.lexicon.get_right_id(0))
        self.assertEqual(8729, self.lexicon.get_cost(0))

        # 東京都
        self.assertEqual(6, self.lexicon.get_left_id(6))
        self.assertEqual(8, self.lexicon.get_right_id(6))
        self.assertEqual(5320, self.lexicon.get_cost(6))

        # 都
        self.assertEqual(8, self.lexicon.get_left_id(9))
        self.assertEqual(8, self.lexicon.get_right_id(9))
        self.assertEqual(2914, self.lexicon.get_cost(9))

    def test_wordinfo(self):
        # た
        wi = self.lexicon.get_word_info(0)
        self.assertEqual('た', wi.surface)
        self.assertEqual(3, wi.head_word_length)
        self.assertEqual(0, wi.pos_id)
        self.assertEqual('た', wi.normalized_form)
        self.assertEqual(-1, wi.dictionary_form_word_id)
        self.assertEqual('た', wi.dictionary_form)
        self.assertEqual('タ', wi.reading_form)
        self.assertEqual([], wi.a_unit_split)
        self.assertEqual([], wi.b_unit_split)
        self.assertEqual([], wi.word_structure)

        # 行っ
        wi = self.lexicon.get_word_info(8)
        self.assertEqual('行っ', wi.surface)
        self.assertEqual('行く', wi.normalized_form)
        self.assertEqual(7, wi.dictionary_form_word_id)
        self.assertEqual('行く', wi.dictionary_form)

        # 東京都
        wi = self.lexicon.get_word_info(6)
        self.assertEqual('東京都', wi.surface)
        self.assertEqual([5, 9], wi.a_unit_split)
        self.assertEqual([], wi.b_unit_split)
        self.assertEqual([5, 9], wi.word_structure)
        self.assertEqual([], wi.synonym_group_ids)

    def test_wordinfo_with_longword(self):
        # 0123456789 * 30
        wi = self.lexicon.get_word_info(36)
        self.assertEqual(300, len(wi.surface))
        self.assertEqual(300, wi.head_word_length)
        self.assertEqual(300, len(wi.normalized_form))
        self.assertEqual(-1, wi.dictionary_form_word_id)
        self.assertEqual(300, len(wi.dictionary_form))
        self.assertEqual(570, len(wi.reading_form))

    def test_size(self):
        self.assertEqual(39, self.lexicon.size())


if __name__ == '__main__':
    unittest.main()
